# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""prefpicker module main"""

from argparse import ArgumentParser
from logging import DEBUG, INFO, basicConfig, getLogger
from os import getenv
from pathlib import Path

from .prefpicker import PrefPicker, SourceDataError

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


def parse_args(argv=None):  # pylint: disable=missing-docstring
    parser = ArgumentParser(description="PrefPicker - Manage & generate prefs.js files")
    parser.add_argument(
        "input",
        type=Path,
        help="Template containing definitions. This can be the path "
        "to a template (YAML) file or the name of a built-in template. "
        "Built-in templates: %s" % (", ".join(x.name for x in PrefPicker.templates()),),
    )
    parser.add_argument("output", type=Path, help="Path of prefs.js file to create.")
    parser.add_argument(
        "--check", action="store_true", help="Display output of sanity checks."
    )
    parser.add_argument("--variant", default="default", help="Specify variant to use.")
    args = parser.parse_args(argv)
    # handle using built-in templates
    builtin_template = PrefPicker.lookup_template(args.input.name)
    if builtin_template:
        args.input = builtin_template
    elif not args.input.is_file():
        parser.error(f"Cannot find input file '{args.input}'")
    # sanity check output
    if args.output.is_dir():
        parser.error(f"Output '{args.output}' is a directory.")
    if not args.output.parent.is_dir():
        parser.error(f"Output '{args.output.parent}' directory does not exist.")
    return args


def main(argv=None):  # pylint: disable=missing-docstring
    if bool(getenv("DEBUG")):  # pragma: no cover
        log_fmt = "%(asctime)s %(levelname).1s %(name)s | %(message)s"
        log_level = DEBUG
    else:
        log_fmt = "%(message)s"
        log_level = INFO
    basicConfig(format=log_fmt, level=log_level)

    args = parse_args(argv)

    LOG.info("Loading %r...", args.input.name)
    try:
        pick = PrefPicker.load_template(args.input)
    except SourceDataError as exc:
        LOG.error("Failed to load '%s': %s", args.input, exc)
        return 1
    LOG.info("Loaded %d prefs and %d variants", len(pick.prefs), len(pick.variants))
    if args.check:
        for result in pick.check_combinations():
            LOG.info(
                "Check: %r variant has %r possible combination(s)", result[0], result[1]
            )
        for result in pick.check_overwrites():
            LOG.info(
                "Check: %r variant %r redefines value %r (may be intentional)",
                result[0],
                result[1],
                result[2],
            )
        for result in pick.check_duplicates():
            LOG.info(
                "Check: %r variant %r contains duplicate values", result[0], result[1]
            )
    if args.variant not in pick.variants:
        LOG.error("Error: Variant %r does not exist", args.variant)
        return 1
    LOG.info("Generating %r using variant %r...", args.output.name, args.variant)
    pick.create_prefsjs(args.output, args.variant)
    LOG.info("Done.")
    return 0



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""prefpicker module"""

from datetime import datetime
from hashlib import sha1
from pathlib import Path
from random import choice

from yaml import safe_load
from yaml.parser import ParserError
from yaml.scanner import ScannerError

__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


class SourceDataError(Exception):
    """This is raised when issues are found in the source data."""


class PrefPicker:  # pylint: disable=missing-docstring

    __slots__ = ("prefs", "variants")

    def __init__(self):
        self.prefs = dict()
        self.variants = {"default"}

    def check_combinations(self):
        """Count the number of combinations for each variation. Only return
           variants that have more than one combination.

        Args:
            None

        Yields:
            tuple: variant and number of potential combinations
        """
        combos = dict.fromkeys(self.variants, 1)
        for variants in (x["variants"] for x in self.prefs.values()):
            for variant in combos:
                # use 'default' if pref does not have a matching variant entry
                if variant not in variants:
                    combos[variant] *= len(variants["default"])
                else:
                    combos[variant] *= len(variants[variant])
        for variant, count in sorted(combos.items()):
            if count > 1:
                yield (variant, count)

    def check_duplicates(self):
        """Look for variants with values that appear more than once per variant.

        Args:
            None

        Yields:
            tuple: pref and the variant
        """
        for pref, keys in sorted(self.prefs.items()):
            variants = keys["variants"]
            for variant in variants:
                if len(variants[variant]) != len(set(variants[variant])):
                    yield (pref, variant)

    def check_overwrites(self):
        """Look for variants that overwrite the default with the same value.

        Args:
            None

        Yields:
            tuple: pref, variant and the value
        """
        for pref, keys in sorted(self.prefs.items()):
            variants = keys["variants"]
            for variant in variants:
                if variant == "default":
                    continue
                for value in variants[variant]:
                    if value in variants["default"]:
                        yield (pref, variant, value)

    def create_prefsjs(self, dest, variant="default"):
        """Write a `prefs.js` file based on the specified variant. The output file
           will also include comments containing the variant, a timestamp and a
           fingerprint. The fingerprint is a hash of pref/value pairs which can
           be used to help catch different files without a diff.

        Args:
            dest (Path): Name including path of file to create.
            variant (str): Used to pick the values to output.

        Returns:
            None
        """
        # create a fingerprint based on prefs/values combinations
        uid = sha1()
        with dest.open("w") as prefs_fp:
            prefs_fp.write("// Generated with PrefPicker @ ")
            prefs_fp.write(datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S UTC"))
            prefs_fp.write(f"\n// Variant {variant!r}\n")
            for pref, keys in sorted(self.prefs.items()):
                variants = keys["variants"]
                # choose values
                if variant not in variants or variant == "default":
                    value = choice(variants["default"])
                    default_variant = True
                else:
                    value = choice(variants[variant])
                    default_variant = False
                if value is None:
                    # skipping pref
                    continue
                # sanitize value for writing
                if isinstance(value, bool):
                    sanitized = "true" if value else "false"
                elif isinstance(value, int):
                    sanitized = str(value)
                elif isinstance(value, str):
                    sanitized = repr(value)
                else:
                    prefs_fp.write(f"// Failed to sanitized {value!r} ({pref})\n")
                    raise SourceDataError(
                        f"Unsupported datatype {type(value).__name__!r}"
                    )
                # write to prefs.js file
                if not default_variant:
                    prefs_fp.write(f"// {pref!r} defined by variant {variant!r}\n")
                prefs_fp.write(f'user_pref("{pref}", {sanitized});\n')
                # update fingerprint
                uid.update(pref.encode(encoding="utf-8", errors="ignore"))
                uid.update(sanitized.encode(encoding="utf-8", errors="ignore"))
            prefs_fp.write(f"// Fingerprint {uid.hexdigest()!r}\n")

    @classmethod
    def lookup_template(cls, name):
        """Lookup built-in template Path.

        Args:
            Name (str): Name of template.

        Returns:
            Path: Template that matches 'name' or None.
        """
        path = Path(__file__).parent / "templates" / name
        if path.is_file():
            return path
        return None

    @classmethod
    def load_template(cls, input_yml):
        """Load data from a template YAML file.

        Args:
            input_yml (Path): Input file.

        Returns:
            PrefPicker
        """
        try:
            raw_prefs = safe_load(input_yml.read_bytes())
        except (ScannerError, ParserError):
            raise SourceDataError("invalid YAML") from None
        cls.verify_data(raw_prefs)
        picker = cls()
        picker.variants = set(raw_prefs["variant"] + ["default"])
        picker.prefs = raw_prefs["pref"]
        return picker

    @staticmethod
    def templates():
        """Available YAML template files.

        Args:
            None

        Yields:
            Path: Template file.
        """
        path = Path(__file__).parent.resolve() / "templates"
        if path.is_dir():
            for template in path.iterdir():
                if template.suffix.lower().endswith(".yml"):
                    yield template

    @staticmethod
    def verify_data(raw_data):
        """Perform strict sanity checks on raw_data. This exists to help prevent
           the template file from breaking or becoming unmaintainable.

        Args:
            raw_data (dict): Data to verify.

        Returns:
            None
        """
        # check variant list
        if "variant" not in raw_data:
            raise SourceDataError("variant list is missing")
        if not isinstance(raw_data["variant"], list):
            raise SourceDataError("variant is not a list")
        # check variant list entries
        valid_variants = {"default"}
        for variant in raw_data["variant"]:
            if not isinstance(variant, str):
                raise SourceDataError("variant definition must be a string")
            valid_variants.add(variant)
        # check prefs dict
        if "pref" not in raw_data:
            raise SourceDataError("pref group is missing")
        if not isinstance(raw_data["pref"], dict):
            raise SourceDataError("pref is not a dict")
        # check entries in prefs dict
        used_variants = set()
        for pref, keys in raw_data["pref"].items():
            if not isinstance(keys, dict):
                raise SourceDataError(f"{pref!r} entry must contain a dict")
            variants = keys.get("variants")
            if not isinstance(variants, dict):
                raise SourceDataError(f"{pref!r} is missing 'variants' dict")
            if "default" not in variants:
                raise SourceDataError(f"{pref!r} is missing 'default' variant")
            # verify variants
            for variant in variants:
                if variant not in valid_variants:
                    raise SourceDataError(
                        f"{variant!r} in {pref!r} is not a defined variant"
                    )
                if not isinstance(variants[variant], list):
                    raise SourceDataError(
                        f"variant {variant!r} in {pref!r} is not a list"
                    )
                if not variants[variant]:
                    raise SourceDataError(f"{variant!r} in {pref!r} is empty")
                for value in variants[variant]:
                    if value is not None and not isinstance(value, (bool, int, str)):
                        raise SourceDataError(
                            f"unsupported datatype {type(value).__name__!r} ({pref})"
                        )
                used_variants.add(variant)
        if valid_variants - used_variants:
            raise SourceDataError(
                f"Unused variants {' '.join(valid_variants - used_variants)!r}"
            )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""prefpicker.py tests"""

from pytest import mark, raises

from .prefpicker import PrefPicker, SourceDataError


def test_prefpicker_01(tmp_path):
    """test simple PrefPicker"""
    yml = tmp_path / "test.yml"
    yml.write_text(
        """
        variant: []
        pref:
          test.a:
            variants:
              default: [1]"""
    )
    picker = PrefPicker.load_template(yml)
    assert len(picker.variants) == 1
    assert "default" in picker.variants
    assert len(picker.prefs) == 1
    assert "test.a" in picker.prefs


@mark.parametrize(
    "data, msg",
    [
        # variant list missing
        ({"pref": {"a.b": {"variants": {"default": [1]}}}}, "variant list is missing"),
        # variant definition is invalid type
        ({"variant": [{"bad": 1}]}, "variant definition must be a string"),
        # variant is invalid type
        ({"variant": ""}, "variant is not a list"),
        # pref dict missing
        ({"variant": []}, "pref group is missing"),
        # pref is invalid type
        ({"pref": [], "variant": []}, "pref is not a dict"),
        # pref entry is invalid
        ({"pref": {"a.b": None}, "variant": []}, "'a.b' entry must contain a dict"),
        # pref variants is invalid
        (
            {"pref": {"a.b": {"variants": None}}, "variant": []},
            "'a.b' is missing 'variants' dict",
        ),
        # pref missing default variant
        (
            {"variant": [], "pref": {"test.a": {"variants": {}}}},
            "'test.a' is missing 'default' variant",
        ),
        # template with undefined variant
        (
            {"variant": [], "pref": {"a.b": {"variants": {"default": [1], "x": [2]}}}},
            "'x' in 'a.b' is not a defined variant",
        ),
        # template with unused variant
        (
            {"variant": ["unused"], "pref": {"a.b": {"variants": {"default": [1]}}}},
            "Unused variants 'unused'",
        ),
        # pref with empty variant list
        (
            {"variant": [], "pref": {"a.b": {"variants": {"default": []}}}},
            "'default' in 'a.b' is empty",
        ),
        # pref with invalid variant type
        (
            {"variant": [], "pref": {"a.b": {"variants": {"default": "x"}}}},
            "variant 'default' in 'a.b' is not a list",
        ),
        # pref variant with invalid type
        (
            {"variant": [], "pref": {"a.b": {"variants": {"default": [1.11]}}}},
            "unsupported datatype 'float' \\(a.b\\)",
        ),
    ],
)
def test_prefpicker_02(data, msg):
    """test PrefPicker.verify_data()"""
    with raises(SourceDataError, match=msg):
        PrefPicker.verify_data(data)


def test_prefpicker_03():
    """test PrefPicker.check_overwrites()"""
    raw_data = {
        "variant": ["fail", "safe"],
        "pref": {
            "test.a": {"variants": {"default": [1], "fail": [1, 2], "safe": [3]}},
            "test.b": {"variants": {"default": [9], "safe": [None]}},
        },
    }
    # use verify_data just to sanity check test data
    PrefPicker.verify_data(raw_data)
    ppick = PrefPicker()
    ppick.variants = set(raw_data["variant"] + ["default"])
    ppick.prefs = raw_data["pref"]
    results = tuple(ppick.check_overwrites())
    assert results
    assert results[0][0] == "test.a"
    assert results[0][1] == "fail"
    assert results[0][2] == 1


def test_prefpicker_04():
    """test PrefPicker.check_duplicates()"""
    raw_data = {
        "variant": ["fail", "safe"],
        "pref": {
            "test.a": {"variants": {"default": [1], "fail": [1, 2, 3, 1], "safe": [3]}},
            "test.b": {"variants": {"default": [9], "safe": [None]}},
        },
    }
    # use verify_data just to sanity check test data
    PrefPicker.verify_data(raw_data)
    ppick = PrefPicker()
    ppick.variants = set(raw_data["variant"] + ["default"])
    ppick.prefs = raw_data["pref"]
    results = tuple(ppick.check_duplicates())
    assert results
    assert results[0][0] == "test.a"
    assert results[0][1] == "fail"


def test_prefpicker_05():
    """test PrefPicker.check_combinations()"""
    raw_data = {
        "variant": ["v1", "v2"],
        "pref": {
            "test.a": {"variants": {"default": [1], "v1": [1, 2, 3, 4], "v2": [3]}},
            "test.b": {"variants": {"default": [2], "v1": [1, 2], "v2": [3]}},
            "test.c": {"variants": {"default": [1, 3], "v2": [None]}},
        },
    }
    # use verify_data just to sanity check test data
    PrefPicker.verify_data(raw_data)
    ppick = PrefPicker()
    ppick.variants = set(raw_data["variant"] + ["default"])
    ppick.prefs = raw_data["pref"]
    results = tuple(ppick.check_combinations())
    assert len(results) == 2
    assert results[0][0] == "default"
    assert results[0][1] == 2
    assert results[1][0] == "v1"
    assert results[1][1] == 16


def test_prefpicker_06(tmp_path):
    """test simple PrefPicker.create_prefsjs()"""
    ppick = PrefPicker()
    prefs = tmp_path / "prefs.js"
    ppick.create_prefsjs(prefs)
    assert prefs.is_file()
    # check only comments were written to the document
    with prefs.open("r") as in_fp:
        for line in in_fp:
            assert line.startswith("//")
        assert in_fp.tell() > 0


def test_prefpicker_07(tmp_path):
    """test PrefPicker.create_prefsjs() with variants"""
    raw_data = {
        "variant": ["test", "skip"],
        "pref": {
            "test.a": {"variants": {"default": [0], "test": [1], "skip": [2]}},
            "test.b": {"variants": {"default": [True]}},
        },
    }
    PrefPicker.verify_data(raw_data)
    ppick = PrefPicker()
    ppick.variants = set(raw_data["variant"] + ["default"])
    ppick.prefs = raw_data["pref"]
    prefs = tmp_path / "prefs.js"
    # test with 'default' variant
    ppick.create_prefsjs(prefs)
    assert prefs.is_file()
    prefs_data = prefs.read_text()
    assert "defined by variant 'default'" not in prefs_data
    assert 'user_pref("test.a", 0);\n' in prefs_data
    assert 'user_pref("test.b", true);\n' in prefs_data
    # test with 'test' variant
    ppick.create_prefsjs(prefs, variant="test")
    assert prefs.is_file()
    prefs_data = prefs.read_text()
    assert 'user_pref("test.a", 1);\n' in prefs_data
    assert 'user_pref("test.b", true);\n' in prefs_data
    assert "// 'test.a' defined by variant 'test'" in prefs_data


def test_prefpicker_08(tmp_path):
    """test PrefPicker.create_prefsjs() with different values"""
    raw_data = {
        "variant": [],
        "pref": {
            "test.a": {
                "variants": {
                    "default": [0, 1],
                }
            },
            "test.b": {
                "variants": {
                    "default": [None],
                }
            },
            "test.c": {
                "variants": {
                    "default": ["test string"],
                }
            },
            "test.d": {
                "variants": {
                    "default": ["'test' \"string\""],
                }
            },
            "test.e": {
                "variants": {
                    "default": [True],
                }
            },
            "test.f": {
                "variants": {
                    "default": [False],
                }
            },
        },
    }
    PrefPicker.verify_data(raw_data)
    ppick = PrefPicker()
    ppick.variants = set(raw_data["variant"] + ["default"])
    ppick.prefs = raw_data["pref"]
    prefs = tmp_path / "prefs.js"
    ppick.create_prefsjs(prefs)
    assert prefs.is_file()
    prefs_data = prefs.read_text()
    assert 'user_pref("test.b",' not in prefs_data
    assert "user_pref(\"test.c\", 'test string');" in prefs_data
    # test with unsupported value datatype
    raw_data = {"variant": [], "pref": {"boom.": {"variants": {"default": [1.01]}}}}
    ppick.variants = set(raw_data["variant"] + ["default"])
    ppick.prefs = raw_data["pref"]
    with raises(SourceDataError, match="Unsupported datatype"):
        ppick.create_prefsjs(prefs)


def test_prefpicker_09(tmp_path):
    """test PrefPicker.load_template() with invalid YAML"""
    yml = tmp_path / "test.yml"
    yml.write_text("{-{-{-{-:::")
    with raises(SourceDataError, match=r"invalid YAML"):
        PrefPicker.load_template(yml)


def test_prefpicker_10():
    """test PrefPicker.lookup_template()"""
    # unknown template
    assert PrefPicker.lookup_template("missing") is None
    # existing template
    template = tuple(PrefPicker.templates())[0]
    assert template
    assert PrefPicker.lookup_template(template.name)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""main.py tests"""

from pytest import raises

from .main import main
from .prefpicker import PrefPicker


def test_main_01(tmp_path):
    """test main()"""
    prefs_js = tmp_path / "prefs.js"
    yml = tmp_path / "test.yml"
    yml.write_text(
        """
        variant: []
        pref:
          test.a:
            variants:
              default: [1]"""
    )
    assert main([str(yml), str(prefs_js), "--variant", "default"]) == 0
    assert prefs_js.is_file()


def test_main_02(capsys):
    """test main() with missing input"""
    with raises(SystemExit):
        main(["missing.yml", "prefs.js"])
    assert "Cannot find input file 'missing.yml'" in capsys.readouterr()[1]


def test_main_03(tmp_path):
    """test main() with built-in input"""
    prefs_js = tmp_path / "prefs.js"
    templates = tuple(x.name for x in PrefPicker.templates())
    assert templates
    assert main([templates[0], str(prefs_js)]) == 0
    assert prefs_js.is_file()


def test_main_04(tmp_path):
    """test main() with invalid variant"""
    prefs_js = tmp_path / "prefs.js"
    yml = tmp_path / "test.yml"
    yml.write_text(
        """
        variant: []
        pref:
          test.a:
            variants:
              default: [1]"""
    )
    assert main([str(yml), str(prefs_js), "--variant", "x"]) == 1
    assert not prefs_js.is_file()


def test_main_05(tmp_path):
    """test main() with check results"""
    prefs_js = tmp_path / "prefs.js"
    yml = tmp_path / "test.yml"
    yml.write_text(
        """
        variant: [extra]
        pref:
          test.a:
            variants:
              default: [1, 1]
              extra: [1, 1]"""
    )
    assert main([str(yml), str(prefs_js), "--check", "--variant", "extra"]) == 0
    assert prefs_js.is_file()


def test_main_06(capsys, tmp_path):
    """test main() with invalid output path"""
    yml = tmp_path / "test.yml"
    yml.touch()
    # output is a directory
    with raises(SystemExit):
        main([str(yml), str(tmp_path)])
    assert "is a directory." in capsys.readouterr()[1]
    # output to missing directory
    with raises(SystemExit):
        main([str(yml), str(tmp_path / "missing" / "prefs.js")])
    assert "directory does not exist." in capsys.readouterr()[1]


def test_main_07(tmp_path):
    """test main() with invalid input"""
    yml = tmp_path / "test.yml"
    yml.write_text("{test{")
    assert main([str(yml), str(tmp_path / "prefs.js")]) == 1



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""template.py tests"""

from difflib import unified_diff

from yaml import safe_dump, safe_load

from .main import main
from .prefpicker import PrefPicker


def test_templates_01(tmp_path):
    """sanity check template YAML files"""
    prefs_js = tmp_path / "prefs.js"
    checked = []
    for template in PrefPicker.templates():
        assert main([str(template), str(prefs_js), "--check"]) == 0
        assert prefs_js.is_file()
        prefs_js.unlink()
        checked.append(template.name)
    assert "browser-fuzzing.yml" in checked


def test_templates_02():
    """check formatting of template YAML files"""
    for template in PrefPicker.templates():
        with template.open() as in_fp:
            # remove comments
            input_yml = "".join(x for x in in_fp if not x.lstrip().startswith("#"))
        formatted_yml = safe_dump(safe_load(input_yml), indent=2, width=100)
        diff = tuple(
            unified_diff(
                input_yml.splitlines(),
                formatted_yml.splitlines(),
                fromfile=str(template),
                tofile="formatting fixes",
                lineterm="",
            )
        )
        assert not diff, "Formatting changes required:\n{}".format("\n".join(diff))



#!/usr/bin/env python
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.
"""setuptools install script"""

import site

from setuptools import setup

site.ENABLE_USER_SITE = True


if __name__ == "__main__":
    setup()
